from copy import deepcopy
from ftw.builder import Builder
from ftw.builder import create
from izug.ticketbox.browser.helper import map_attribute
from izug.ticketbox.testing import TICKETBOX_INTEGRATION_TESTING
from izug.ticketbox.tests import helpers
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from Products.Archetypes.event import ObjectEditedEvent
from Products.CMFCore.utils import getToolByName
from unittest2 import TestCase
from zope.component import getMultiAdapter
from zope.event import notify


def get_ids_and_titles(data):
    """Returns all ids and titles of the dicts within the list data.
    """
    return [(item.get('id'), item.get('title')) for item in data]


class TestTicketBox(TestCase):

    layer = TICKETBOX_INTEGRATION_TESTING

    def setUp(self):
        super(TestTicketBox, self).setUp()
        self.portal = self.layer['portal']
        self.wf_tool = getToolByName(self.portal, 'portal_workflow')
        self.wf_tool.setDefaultChain('simple_publication_workflow')
        helpers.login_as_manager(self.portal)

        self.catalog = getToolByName(self.portal, 'portal_catalog')

        self.ticketbox = helpers.create_ticketbox(self.portal)
        self.ticket1 = helpers.create_ticket(self.ticketbox)
        self.ticket2 = helpers.create_ticket(self.ticketbox, data_index=1)

    def tearDown(self):
        helpers.logout_manager(self.portal)
        super(TestTicketBox, self).tearDown()

    def test_required_fields(self):
        title = self.ticketbox.getField('title')
        description = self.ticketbox.getField('description')

        self.assertEquals(title.required, True)
        self.assertEquals(description.required, False)

    def test_generated_id(self):
        state = self.ticketbox.getAvailableStates()[2]['id']
        area = self.ticketbox.getAvailableAreas()[2]['id']
        release = self.ticketbox.getAvailableReleases()[2]['id']
        priority = self.ticketbox.getAvailablePriorities()[2]['id']

        self.assertEquals(state, "test-id-state")
        self.assertEquals(area, "test-id-area")
        self.assertEquals(release, "test-id-release")
        self.assertEquals(priority, "test-id-priority")

    def test_overview_view(self):

        def get_filtered_tickets(filterid, filtervalue):
            self.portal.REQUEST.set('filterid', filterid)
            self.portal.REQUEST.set('filtervalue', filtervalue)
            view = getMultiAdapter(
                (self.ticketbox, self.portal.REQUEST),
                name='tabbedview_view-overview')
            view.update()
            contents = view.contents
            self.portal.REQUEST.set('filterid', None)
            self.portal.REQUEST.set('filtervalue', None)
            return contents

        ticket_state = get_filtered_tickets('getState', 'show-all-tickets')
        ticket_area = get_filtered_tickets('getArea', 'internet')
        ticket_priority = get_filtered_tickets('getPriority', 'high')
        ticket_responsible = get_filtered_tickets('getResponsibleManager',
                                                  'test_user_1_')

        self.assertEquals(ticket_state[0].getObject().getState(), "show-all-tickets")
        self.assertEquals(ticket_area[0].getObject().getArea(), "internet")
        self.assertEquals(ticket_priority[0].getObject().getPriority(),
                          "high")
        self.assertEquals(
            ticket_responsible[0].getObject().getResponsibleManager(),
            "test_user_1_")

    def test_all_tickets_view(self):
        ticketbox_view = getMultiAdapter(
            (self.ticketbox, self.portal.REQUEST),
            name='tabbedview_view-all_tickets')
        ticketbox_view.update()

        self.assertEquals(len(ticketbox_view.contents), 1)

    def test_my_tickets_view(self):
        ticketbox_view = getMultiAdapter(
            (self.ticketbox, self.portal.REQUEST),
            name='tabbedview_view-my_tickets')
        ticketbox_view.update()

        self.assertEquals(len(ticketbox_view.contents), 1)

    def test_unassigned_is_assignable(self):
        self.assertIn(
            ('(UNASSIGNED)', u'None'),
            self.ticketbox.assignable_users())

    def test_make_user_not_assignable(self):
        setRoles(self.portal, TEST_USER_ID, ['Member'])
        field = self.ticketbox.getField('assignable_user_ids')
        mutator = field.getMutator(self.ticketbox)

        self.ticketbox.manage_setLocalRoles(
            'test_user_1_', ['Contributor', 'Reader'])
        self.assertIn(
            ('test_user_1_', 'test-user'),
            self.ticketbox.assignable_users())

        mutator([])
        self.assertNotIn(
            ('test_user_1_', 'test-user'),
            self.ticketbox.assignable_users())

        mutator(['test_user_1_'])
        self.assertIn(
            ('test_user_1_', 'test-user'),
            self.ticketbox.assignable_users())

    def test_new_user_gets_assignable_automatically(self):
        user1 = create(Builder('user'))
        #You should never set localroles on a plonesite but we do it here since
        # it is our only parent and we do require local roles.
        self.portal.manage_setLocalRoles(
            'john.doe', ['Contributor', 'Reader'])
        self.assertIn(
            ('john.doe', 'Doe John'),
            self.ticketbox.assignable_users())

    def test_only_with_add_permission(self):
        user1 = create(Builder('user'))
        user2 = create(Builder('user').named('James','Bond'))
        self.ticketbox.manage_setLocalRoles(
            user1.id, ['Contributor', 'Reader'])
        self.ticketbox.manage_setLocalRoles(
            user2.id, ['Reader'])

        self.assertIn(
            ('john.doe', 'Doe John'),
            self.ticketbox.assignable_users())

        self.assertNotIn(
            ('james.bond', 'Bond James'),
            self.ticketbox.assignable_users())

    def test_duplicate_state_ids(self):
        # When changing a state and adding a state in a specific way it
        # should not result in duplicate state ids.

        states = [{'id': '',
                   'title': 'Open',
                   'show_in_all_tickets': '1',
                   'show_in_my_tickets': '1'}]
        self.ticketbox.setAvailableStates(states)
        notify(ObjectEditedEvent(self.ticketbox))

        states = self.ticketbox.getAvailableStates()
        self.assertEqual(len(states), 1)
        self.assertEqual(get_ids_and_titles(states), [('open', 'Open')])

        states = list(deepcopy(states))
        states.append({'id': '',
                       'title': 'Open',
                       'show_in_all_tickets': '1',
                       'show_in_my_tickets': '1'})
        self.ticketbox.setAvailableStates(states)
        notify(ObjectEditedEvent(self.ticketbox))

        states = self.ticketbox.getAvailableStates()
        self.assertEqual(len(states), 2)
        self.assertEqual(get_ids_and_titles(states), [
                ('open', 'Open'),
                ('open-1', 'Open')])

    def test_duplicate_release_ids(self):
        releases = [{'id': '',
                     'title': '1.x'}]
        self.ticketbox.setAvailableReleases(releases)
        notify(ObjectEditedEvent(self.ticketbox))

        releases = self.ticketbox.getAvailableReleases()
        self.assertEqual(len(releases), 1)
        self.assertEqual(get_ids_and_titles(releases), [('1-x', '1.x')])

        releases = list(deepcopy(releases))
        releases.append({'id': '',
                         'title': '1.x'})
        self.ticketbox.setAvailableReleases(releases)
        notify(ObjectEditedEvent(self.ticketbox))

        releases = self.ticketbox.getAvailableReleases()
        self.assertEqual(len(releases), 2)
        self.assertEqual(get_ids_and_titles(releases),
                         [('1-x', '1.x'),
                          ('1-x-1', '1.x')])

    def test_sorted_tickets_after_changing_datagrid_titles(self):
        STATE_DATA = [
            {'id': '',
             'title': "A state",
             'show_in_all_tickets': '1',
             'show_in_my_tickets': '0',
             },

            {'id': '',
             'title': "B state",
             'show_in_all_tickets': '0',
             'show_in_my_tickets': '1',
             },

            {'id': '',
             'title': "C state",
             'show_in_all_tickets': '1',
             'show_in_my_tickets': '1',
             }]

        # Set new states to the ticketbox and let it generate the
        # ids for each state by notify the edited-event.
        self.ticketbox.setAvailableStates(STATE_DATA)
        notify(ObjectEditedEvent(self.ticketbox))

        states = self.ticketbox.getAvailableStates()

        # Use the new states for the tickets
        self.ticket1.setState(states[2].get('id'))
        self.ticket1.setTitle('Ticket 1')
        self.ticket1.reindexObject()

        self.ticket2.setState(states[0].get('id'))
        self.ticket2.setTitle('Ticket 2')
        self.ticket2.reindexObject()

        query = {'portal_type': 'Ticket', 'sort_on': 'getState'}

        # Check the sorting by state
        tickets = [ticket.getObject() for ticket in self.catalog(query)]
        self.assertEqual(
            [['A state', 'Ticket 2'], ['C state', 'Ticket 1']],
            [[map_attribute(ticket, "state"), ticket.Title()] for ticket in tickets])

        # Rename the title of the first state in the ticketbox...
        states[0]['title'] = 'D state'
        notify(ObjectEditedEvent(self.ticketbox))

        # And check the sort-order of the tickets again
        tickets = [ticket.getObject() for ticket in self.catalog(query)]
        self.assertEqual(
            [['C state', 'Ticket 1'], ['D state', 'Ticket 2']],
            [[map_attribute(ticket, "state"), ticket.Title()] for ticket in tickets],
            "The sort-order should have been changed because we changed the state title "
            "on the ticketbox.")



from ftw.upgrade import UpgradeStep


class AddTicketboxTitleIndex(UpgradeStep):

    def __call__(self):
        self.setup_install_profile('profile-izug.ticketbox.upgrades:4100')
        self.catalog_reindex_objects({'portal_type': 'Ticket'},
                                     idxs=['ticketbox_title'])



from AccessControl import ClassSecurityInfo
from izug.ticketbox.config import PROJECTNAME
from izug.ticketbox.interfaces import IAttachment
from Products.Archetypes import atapi
from Products.Archetypes.BaseContent import BaseContent
from Products.ATContentTypes.config import ICONMAP
from Products.ATContentTypes.content.file import ATFile
from Products.CMFCore.utils import getToolByName
from Products.MimetypesRegistry.common import MimeTypeException
from urllib import quote
from zope.interface import implements
import logging


class TicketAttachment(ATFile):
    """A file content type based on blobs.
    """
    implements(IAttachment)
    security = ClassSecurityInfo()

    security.declarePublic('getIcon')
    def getIcon(self, relative_to_portal=0):
        """Calculate the icon using the mime type of the file
        """

        field = self.getField('file')
        if not field or not self.get_size():
            # field is empty
            return BaseContent.getIcon(self, relative_to_portal)

        contenttype = field.getContentType(self)
        contenttype_major = contenttype and contenttype.split('/')[0] or ''

        mtr = getToolByName(self, 'mimetypes_registry', None)
        utool = getToolByName(self, 'portal_url')

        mimetypeitem = None
        try:
            mimetypeitem = mtr.lookup(contenttype)
        except MimeTypeException, msg:
            LOG = logging.getLogger('ATCT')
            LOG.error('MimeTypeException for {0}. Error is: {1}'.format(
                self.absolute_url(),
                str(msg)))
        if not mimetypeitem:
            icon = None
        else:
            icon = mimetypeitem[0].icon_path

        if not icon:
            if contenttype in ICONMAP:
                icon = quote(ICONMAP[contenttype])
            elif contenttype_major in ICONMAP:
                icon = quote(ICONMAP[contenttype_major])
            else:
                return BaseContent.getIcon(self, relative_to_portal)

        if relative_to_portal:
            return icon
        else:
            # Relative to REQUEST['BASEPATH1']
            res = utool(relative=1) + '/' + icon
            while res[:1] == '/':
                res = res[1:]
            return res

atapi.registerType(TicketAttachment, PROJECTNAME)



from ftw.upgrade import ProgressLogger
from ftw.upgrade import UpgradeStep
import logging


LOG = logging.getLogger('izug.ticketbox.upgrades')


class MigrateTicketDescription(UpgradeStep):

    def __call__(self):
        catalog = self.getToolByName('portal_catalog')
        brains = catalog.unrestrictedSearchResults(
            portal_type='Ticket')

        title = 'Migrate ticket description to rich text field'
        with ProgressLogger(title, brains) as step:

            for brain in brains:
                self._migrate_description(brain)
                step()

    def _migrate_description(self, brain):
        obj = self.portal.unrestrictedTraverse(brain.getPath())

        field = obj.getField('description')
        field.set(obj, field.getRaw(obj))



from Acquisition import aq_inner, aq_parent
from Products.CMFCore.utils import getToolByName
from ftw.tabbedview.browser.listing import CatalogListingView
from ftw.tabbedview.browser.tabbed import TabbedView
from ftw.table import helper
from izug.ticketbox import ticketboxMessageFactory as _
from izug.ticketbox.browser.tabbed.base import BaseTicketListingTab
import os.path


def link_to_parent(item, value):
    """Tabbedview helper for linking the parent.
    """

    if isinstance(value, str):
        value = value.decode('utf-8')

    parent = aq_parent(aq_inner(item.getObject()))
    title = parent.Description().decode('utf-8')

    return u'<a href="%s" title="%s">%s</a>' % (
        os.path.dirname(item.getURL()),
        title,
        value)


class TicketboxesTabbedView(TabbedView):

    def get_tabs(self):
        return (
            {'id': _('global-my_tickets'),
             'class': ''},

            {'id': _('global-my_issued_tickets'),
             'class': ''},

            {'id': _('global-all_tickets'),
             'class': ''},

            {'id': _('global-ticketboxes'),
             'class': ''},
        )


class GlobalTicketboxesTab(CatalogListingView):

    types = ['Ticket Box']
    show_selects = False
    show_menu = False
    sort_on = 'sortable_title'

    columns = (
        {'column': 'Title',
         'column_title': _(u"Title"),
         'sort_index': 'sortable_title',
         'transform': helper.linked},

        {'column': 'modified',
         'column_title': _(u"ModificationDate"),
         'transform': helper.readable_date},

        {'column': 'get_owner_index',
         'column_title': _(u"Responsible"),
         'transform': helper.readable_author},
    )

    def get_base_query(self):
        site = getToolByName(self.context, 'portal_url').getPortalObject()
        site_path = '/'.join(site.getPhysicalPath())

        query = super(GlobalTicketboxesTab, self).get_base_query()
        query['path']['query'] = site_path
        return query


class GlobalTicketTabBase(BaseTicketListingTab):

    @property
    def columns(self):
        columns = list(super(GlobalTicketTabBase, self).columns)
        columns.append({'column': 'ticketbox_title',
                        'column_title': _(u'Ticket Box'),
                        'transform': link_to_parent})
        return tuple(columns)

    def get_base_query(self):
        site = getToolByName(self.context, 'portal_url').getPortalObject()
        site_path = '/'.join(site.getPhysicalPath())

        query = super(GlobalTicketTabBase, self).get_base_query()
        query['path']['query'] = site_path
        return query

    def _get_ticketbox_path_for(self, item):
        path = item.getPath()
        path, _ticketid = os.path.split(path)
        return path


class GlobalAllTicketsTab(GlobalTicketTabBase):
    """Global tickets tab listing all (visible) tickets.
    """


class GlobalMyTicketsTab(GlobalTicketTabBase):
    """Tab listing all tickets where the current user is responsible.
    """

    def get_base_query(self):
        query = super(GlobalMyTicketsTab, self).get_base_query()

        member = self.context.restrictedTraverse(
            '@@plone_portal_state').member()
        query['getResponsibleManager'] = member.getId()

        return query


class GlobalMyIssuedTicketsTab(GlobalTicketTabBase):
    """Tab listing all tickets where the current user is the creator.
    """

    def get_base_query(self):
        query = super(GlobalMyIssuedTicketsTab, self).get_base_query()

        member = self.context.restrictedTraverse(
            '@@plone_portal_state').member()
        query['Creator'] = member.getId()

        return query



from Acquisition import aq_inner
from DateTime import DateTime
from izug.ticketbox import ticketboxMessageFactory as _
from izug.ticketbox.adapters import Response
from izug.ticketbox.browser.helper import map_attribute
from izug.ticketbox.config import DEFAULT_ISSUE_MIME_TYPE
from izug.ticketbox.interfaces import IResponseAdder
from izug.ticketbox.interfaces import IResponseContainer
from OFS.Image import File
from plone.i18n.normalizer.interfaces import IIDNormalizer
from plone.memoize.view import memoize
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.component import queryUtility
from zope.interface import implements
from zope.lifecycleevent import modified
from zope.schema.interfaces import IVocabularyFactory


class Base(BrowserView):
    """Base view for Ticketbox Response.
    Mostly meant as helper for adding a Ticketbox Response.
    """

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.folder = IResponseContainer(context)
        self.mimetype = DEFAULT_ISSUE_MIME_TYPE
        self.use_wysiwyg = (self.mimetype == 'text/html')

    def responses(self):
        """Returns all Responses in the ticket
        """
        context = aq_inner(self.context)
        trans = context.portal_transforms
        items = []
        for id_, response in enumerate(self.folder):
            if response is None:
                # Has been removed.
                continue
            # Use the already rendered response when available
            if response.rendered_text is None:
                if response.mimetype == 'text/html':
                    html = response.text
                else:
                    html = trans.convertTo('text/html',
                                           response.text,
                                           mimetype=response.mimetype)
                    html = html.getData()
                response.rendered_text = html
            html = response.rendered_text
            info = dict(id=id_,
                        response=response,
                        attachment=self.attachment_info(id_),
                        references=self.references_info(id_),
                        html=html)
            items.append(info)
        return items

    def get_years(self):
        """return a list of years in past and future"""

        years = []
        year = DateTime().year() - 3
        for _i in range(1, 10):
            years.append(year)
            year += 1

        return years

    @property
    @memoize
    def portal_url(self):
        context = aq_inner(self.context)
        plone = context.restrictedTraverse('@@plone_portal_state')
        return plone.portal_url()

    def attachment_info(self, id_):
        """Return TicketAttachment object
        """
        context = aq_inner(self.context)
        response = self.folder[id_]
        # Get attachment uid
        # In future this could be a list of attachments
        attachment_uid = response.attachment
        if attachment_uid is None:
            return None

        return context.reference_catalog.lookupObject(attachment_uid)

    def references_info(self, id_):
        """Return references of response"""
        context = aq_inner(self.context)
        response = self.folder[id_]

        # Fallback for old responses
        if not hasattr(response, 'references'):
            return []

        objs = []
        for uid in response.references:
            if uid:
                obj = context.reference_catalog.lookupObject(uid)
                if obj:
                    objs.append(obj)
        return objs

    def mock_reference_field(self):
        """Mocks reference field for response"""

        field = self.context.getField('ticketReferences')
        startuppath = field.widget.getStartupDirectory(self.context, field)

        class Dummy(object):
            # Disable zope security
            __allow_access_to_unprotected_subobjects__ = True
            multiValued = 1
            # Use the same fieldname for our repsonse widet as we
            # use for the ticket reference (load some settings)
            getName = 'ticketReferences'

            class Widget(object):
                __allow_access_to_unprotected_subobjects__ = True
                allow_sorting = 0
                getStartupDirectory = lambda x, y, z: startuppath
            widget = Widget()

        return Dummy()

    @Lazy
    def memship(self):
        context = aq_inner(self.context)
        return getToolByName(context, 'portal_membership')

    @property
    @memoize
    def can_edit_response(self):
        context = aq_inner(self.context)
        return self.memship.checkPermission(
            'Modify portal content',
            context)

    @property
    @memoize
    def can_delete_response(self):
        context = aq_inner(self.context)
        return self.memship.checkPermission('Delete objects', context)

    def validate_response_id(self):
        """Validate the response id from the request.

        Return -1 if for example the response id does not exist.
        Return the response id otherwise.

        Side effect: an informative status message is set.
        """
        status = IStatusMessage(self.request)
        response_id = self.request.form.get('response_id', None)
        if response_id is None:
            msg = _(u"msg_no_response", default=u"No Response was selected")
            msg = self.context.translate(msg)
            status.addStatusMessage(msg, type='error')
            return -1
        else:
            try:
                response_id = int(response_id)
            except ValueError:
                msg = _(u"msg_nointeger",
                        default=u"Response id ${response_id} is no integer.",
                        mapping=dict(response_id=response_id))
                msg = self.context.translate(msg)
                status.addStatusMessage(msg, type='error')
                return -1
            if response_id >= len(self.folder):
                msg = _(
                    u"msg_invalid",
                    default=u"The Response ID ${response_id} doesn't exist.",
                    mapping=dict(response_id=response_id))
                msg = self.context.translate(msg)
                status.addStatusMessage(msg, type='error')
                return -1
            else:
                return response_id
        # fallback
        return -1

    @property
    def priority(self):
        context = aq_inner(self.context)
        return context.getPriority()

    @property
    def releases(self):
        context = aq_inner(self.context)
        return context.getReleases()

    @property
    def state(self):
        context = aq_inner(self.context)
        return context.getState()

    @property
    @memoize
    def states_for_display(self):
        factory = getUtility(IVocabularyFactory,
                             name='ticketbox_values_states')
        result = []
        for term in factory(self.context.aq_inner):
            current_state = self.context.getState()
            checked = term.token == current_state
            result.append(
                dict(
                    value=term.token,
                    label=term.title,
                    checked=checked))
        return result

    @property
    def available_states(self):
        return [t['value'] for t in self.states_for_display]

    @property
    def priorities_for_display(self):
        context = self.context.aq_inner
        result = []
        for term in context.getAvailablePriorities():
            current_state = self.context.getPriority()
            checked = term['id'] == current_state
            result.append(
                dict(
                    value=term['id'],
                    label=term['title'],
                    checked=checked))

        return result

    @property
    def areas_for_display(self):
        context = self.context.aq_inner
        result = []
        for term in context.getAvailableAreas():
            current_state = self.context.getArea()
            checked = term['id'] == current_state
            result.append(
                dict(
                    value=term['id'],
                    label=term['title'],
                    checked=checked))

        return result

    @property
    @memoize
    def available_areas(self):
        """Get the available severities for this issue.
        """
        return [t['value'] for t in self.areas_for_display]

    @property
    @memoize
    def varieties_for_display(self):
        context = aq_inner(self.context)
        result = []
        for term in context.getAvailableVarieties():
            current_state = self.context.getVariety()
            checked = term['id'] == current_state
            result.append(
                dict(
                    value=term['id'],
                    label=term['title'],
                    checked=checked))

        return result

    @property
    @memoize
    def available_varieties(self):
        """Get the available varieties for this issue.
        """
        return [t['value'] for t in self.varieties_for_display]

    @property
    def responsibleManager(self):
        context = aq_inner(self.context)
        return context.getResponsibleManager()

    @property
    @memoize
    def available_priorities(self):
        """Get the available severities for this issue.
        """
        return [t['value'] for t in self.priorities_for_display]

    @property
    def releases_for_display(self):
        """Get the releases from the project.

        Usually nothing, unless you use Ticketbox in combination with
        PloneSoftwareCenter.
        """
        result = []
        factory = getUtility(
            IVocabularyFactory,
            name='ticketbox_values_releases')
        for term in factory(self.context.aq_inner):
            current_state = self.context.getReleases()
            checked = term.token == current_state
            result.append(
                dict(
                    value=term.token,
                    label=term.title,
                    checked=checked))
        return result

    @property
    @memoize
    def available_releases(self):
        """Get the releases from the project.
        """
        return [t['value'] for t in self.releases_for_display]

    @property
    def watched_releases_for_display(self):
        """Get the releases from the project.

        Usually nothing, unless you use Ticketbox in combination with
        PloneSoftwareCenter.
        """
        result = []
        factory = getUtility(
            IVocabularyFactory,
            name='ticketbox_values_releases')
        for term in factory(self.context.aq_inner):
            current_state = self.context.getWatchedRelease()
            checked = term.token == current_state
            result.append(
                dict(
                    value=term.token,
                    label=term.title,
                    checked=checked))
        return result

    @property
    @memoize
    def available_watched_releases(self):
        """Get the available watched releases for this issue.
        """
        return [t['value'] for t in self.watched_releases_for_display]

    @property
    def show_target_releases(self):
        """Should the option for selecting a target release be shown?

        There is always at least one option: None.  So only show when
        there is more than one option.
        """
        return len(self.available_releases) > 1

    @property
    def show_watched_releases(self):
        """Should the option for selecting a target release be shown?

        There is always at least one option: None.  So only show when
        there is more than one option.
        """
        return len(self.available_watched_releases) > 1

    @property
    def multiple_priorities(self):
        """Should the option for selecting a target release be shown?

        There is always at least one option: None.  So only show when
        there is more than one option.
        """
        return len(self.context.getAvailablePriorities()) > 1

    @property
    def multiple_states(self):
        """Should the option for selecting a target release be shown?

        There is always at least one option: None.  So only show when
        there is more than one option.
        """
        return len(self.states_for_display) > 1

    @property
    def multiple_areas(self):
        """Should the option for selecting a target area be shown?

        There is always at least one option: None.  So only show when
        there is more than one option.
        """
        return len(self.areas_for_display) > 1

    @property
    def multiple_varieties(self):
        """Should the option for selecting a target variety be shown?

        There is always at least one option: None.  So only show when
        there is more than one option.
        """
        return len(self.varieties_for_display) > 1

    @property
    def managers_for_display(self):
        """Get the tracker managers.
        """
        context = self.context.aq_inner
        users = context.assignable_users()
        result = []
        assignedUser = context.getResponsibleManager()
        for user in users:
            result.append(dict(value=user[0],
                               label=user[1],
                               checked=assignedUser == user[0]))
        return result

    @property
    @memoize
    def available_managers(self):
        """Get the tracker managers.
        """
        # get vocab from issue
        result = []
        for user in self.context.aq_inner.assignable_users():
            result.append(user[0])
        return result

    @property
    @memoize
    def upload_allowed(self):
        """Is the user allowed to upload on attachment?
        """
        return True


class AddForm(Base):
    implements(IResponseAdder)

    def __init__(self, context, request, view):
        super(AddForm, self).__init__(context, request)
        self.__parent__ = view

    def update(self):
        pass

    def render(self):
        # self.template is defined in zcml
        return self.template()


class Create(Base):

    def determine_response_type(self, response):
        """Return a string indicating the type of response this is.
        """
        responseCreator = response.creator
        if responseCreator == '(anonymous)':
            return 'additional'

        issue = aq_inner(self.context)
        if responseCreator == issue.Creator():
            return 'clarification'

        if responseCreator in self.available_managers:
            return 'reply'

        # default:
        return 'additional'

    def __call__(self):
        form = self.request.form
        context = aq_inner(self.context)
        modifiedDate = context.modified()
        response_text = form.get('response', u'')
        new_response = Response(response_text)
        new_response.mimetype = self.mimetype
        new_response.type = self.determine_response_type(new_response)

        issue_has_changed = False
        #Unassigned is no member in portal_membership.
        #So we have to set it manually
        unassigned = _(u'label_unassigned', default=u'unassigned')
        responsible_after = form.get('responsibleManager', u'')
        if responsible_after != context.getResponsibleManager():

            #get ResponsibleManager and member-infos before changes
            responsible_before = context.getResponsibleManager()
            member_before = self.context.portal_membership.getMemberById(
                responsible_before)

            context.setResponsibleManager(responsible_after)

            #get member-infos after changes
            member_after = self.context.portal_membership.getMemberById(
                responsible_after)

            #get fullname from member before changes
            if member_before:
                before = member_before.getProperty(
                    'fullname', responsible_before)
            else:
                before = unassigned

            #get fullname from member after changes
            if member_after:
                after = member_after.getProperty(
                    'fullname', responsible_after)
            else:
                after = unassigned

            new_response.add_change('responsibleManager',
                                    _(u'label_responsibleManager',
                                      default=u"Responsible"),
                                    before, after)
            issue_has_changed = True

        # Answerdate
        answerdate_before = context.getAnswerDate()
        if answerdate_before:
            answerdate_before = answerdate_before.strftime('%d.%m.%Y %H:%M')
        else:
            answerdate_before = ''

        answerdate_after = answerdate_before
        year = int(form.get('answerdate_year', 0))
        month = int(form.get('answerdate_month', 0))
        day = int(form.get('answerdate_day', 0))

        if year and month and day:
            hour = int(form.get('answerdate_hour', 0))
            minute = int(form.get('answerdate_minute', 0))
            answerdate_after = DateTime(
                year, month, day, hour, minute).strftime('%d.%m.%Y %H:%M')

        if answerdate_before != answerdate_after:
            context.setAnswerDate(answerdate_after)
            new_response.add_change('answerDate',
                                    _(u'label_answerdate',
                                      default=u'Answerdate'),
                                    answerdate_before,
                                    answerdate_after)
            issue_has_changed = True

        options = [
            ('priority',
             _(u'label_priority_', default=u"Priority"),
             'available_priorities'),

            ('releases',
             _(u'label_releases', default=u"Target Release"),
             'available_releases'),

            ('state',
             _(u'label_state', default=u"State"),
             'available_states'),

            ('area',
             _(u'label_areas', default=u"Area"),
             'available_areas'),

            ('variety',
             _(u'label_varieties', default=u"Variety"),
             'available_varieties'),

            ('watchedRelease',
             _(u'label_watched_release', default=u"Watched Release"),
             'available_watched_releases'),
            ]
        # Changes that need to be applied to the issue (apart from
        # workflow changes that need to be handled separately).
        changes = {}
        for option, title, vocab in options:
            new = form.get(option, u'')
            if new and new in self.__getattribute__(vocab):
                current = context.__getattribute__(option)
                if current != new:
                    changes[option] = new
                    new_response.add_change(
                        option, title,
                        map_attribute(self.context, option, current),
                        map_attribute(self.context, option, new))
                    issue_has_changed = True

        attachment = form.get('attachment')
        if attachment:
            # Create filename like AT - some Browser delivers the
            # local full path
            filename = attachment.filename
            filename = filename[max(
                    filename.rfind('/'),
                    filename.rfind('\\'),
                    filename.rfind(':')) + 1:]

            # File(id, title, file)
            data = File(filename, filename, attachment)
            if not hasattr(data, 'filename'):
                setattr(data, 'filename', filename)
            # Create TicketAttachment and save the uid in attachment attr of
            # new_response
            new_id = queryUtility(IIDNormalizer).normalize(
                filename.decode('utf-8'))
            if context.get(new_id, None):
                IStatusMessage(context.REQUEST).addStatusMessage(
                    _(u"text_file_exists_error"), type='error')
                context.setAttachment('DELETE_FILE')
                return self.request.response.redirect(context.absolute_url())
            new_file_id = context.invokeFactory(
                type_name="TicketAttachment",
                id=new_id,
                title=filename,
                file=data)
            new_file = context.get(new_file_id, None)
            new_response.attachment = new_file.UID()
            issue_has_changed = True

        references = form.get('ticketReferences')
        if references:
            new_response.references = references
            # Store refs also on Ticket
            self.context.setTicketReferences(
                references + self.context.getRawTicketReferences())

        if len(response_text) == 0 and not issue_has_changed:
            status = IStatusMessage(self.request)
            msg = _(
                u"msg_no_changes",
                default="No response text added and no issue changes made.")
            #
            # msg = self.context.translate(msg)
            status.addStatusMessage(msg, type='error')
        else:
            # Apply changes to issue
            # XXX: CHANGE WORKFLOW - OR CHANGE SECURITYMANAGER
            # We cannot use AT's update method, because of a security check
            # we don't want. Let's set the new values manually.
            # OLD:
            # context.update(**changes)

            # NEW:
            if 'releases' in changes:
                context.setReleases(changes['releases'])
            if 'priority' in changes:
                context.setPriority(changes['priority'])
            if 'area' in changes:
                context.setArea(changes['area'])
            if 'variety' in changes:
                context.setVariety(changes['variety'])
            if 'state' in changes:
                context.setState(changes['state'])
            if 'watchedRelease' in changes:
                context.setWatchedRelease(changes['watchedRelease'])

            # Add response
            catalog_tool = self.context.portal_catalog
            # re-set the modification date -
            # this must be the last modifying access
            context.reindexObject()
            self.folder.add(new_response)
            context.setModificationDate(modifiedDate)
            catalog_tool.catalog_object(context,
                                        '/'.join(context.getPhysicalPath()))
        if form.get('sendNotification', None):
            self.request.response.redirect(
                context.absolute_url() + '/notification_form')
        else:
            self.request.response.redirect(context.absolute_url())


class Edit(Base):

    @property
    @memoize
    def response(self):
        form = self.request.form
        response_id = form.get('response_id', None)
        if response_id is None:
            return None
        try:
            response_id = int(response_id)
        except ValueError:
            return None
        if response_id >= len(self.folder):
            return None
        return self.folder[response_id]

    @property
    def response_found(self):
        return self.response is not None


class Save(Base):

    def __call__(self):
        form = self.request.form
        context = aq_inner(self.context)
        status = IStatusMessage(self.request)

        response_id = form.get('response_id', None)
        if response_id is None:
            msg = _(u"msg_no_response_selected",
                    default=u"No response selected for saving.")
            msg = self.context.translate(msg)
            status.addStatusMessage(msg, type='error')
        elif self.folder[response_id] is None:
            msg = _(u"msg_doesnt_exists",
                    default=u"Response does not exist anymore; \
                    perhaps it was removed by another user.")
            msg = self.context.translate(msg)
            status.addStatusMessage(msg, type='error')

        response = self.folder[response_id]

        member = context.portal_membership.getAuthenticatedMember()
        if not (self.can_edit_response or response.creator == member.getId()):
            msg = _(u"msg_not_restricted",
                    default=u"You are not allowed to edit responses.")
            msg = self.context.translate(msg)
            status.addStatusMessage(msg, type='error')
        else:
            response_text = form.get('response', u'')
            response.text = response_text
            # Remove cached rendered response.
            response.rendered_text = None
            msg = _(u"msg_changes_saved", default="Changes Saved",
                    mapping=dict(response_id=response_id))
            msg = self.context.translate(msg)
            status.addStatusMessage(msg, type='info')
            # Fire event.  We put the context in the descriptions
            # so event handlers can use this fully acquisition
            # wrapped object to do their thing.  Feels like
            # cheating, but it gets the job done.  Arguably we
            # could turn the two arguments around and signal that
            # the issue has changed, with the response in the
            # event descriptions.
            modified(response, context)
        self.request.response.redirect(context.absolute_url())


class Delete(Base):

    def __call__(self):
        context = aq_inner(self.context)
        status = IStatusMessage(self.request)

        if not self.can_delete_response:
            msg = _(u"msg_restricted_delete",
                    default=u"You are not allowed to delete Responses")
            msg = self.context.translate(msg)
            status.addStatusMessage(msg, type='error')
        else:
            response_id = self.request.form.get('response_id', None)
            if response_id is None:
                msg = _(u"msg_no_response_delete",
                        default=u"No response selected for removal.")
                msg = self.context.translate(msg)
                status.addStatusMessage(msg, type='error')
            else:
                try:
                    response_id = int(response_id)
                except ValueError:
                    msg = _(
                        u"msg_nointeger",
                        default=u"Response id ${response_id} is no integer.",
                        mapping=dict(response_id=response_id))
                    msg = self.context.translate(msg)
                    status.addStatusMessage(msg, type='error')
                    self.request.response.redirect(context.absolute_url())
                    return

                if response_id >= len(self.folder):
                    msg = _(u"msg_invalid",
                            default=u"The Response ID ${response_id} "
                            "doesn't exist.",
                            mapping=dict(response_id=response_id))
                    msg = self.context.translate(msg)
                    status.addStatusMessage(msg, type='error')
                else:
                    self.folder.delete(response_id)
                    msg = _(u"msg_removed", default="Removed response.",
                            mapping=dict(response_id=response_id))
                    msg = self.context.translate(msg)
                    status.addStatusMessage(msg, type='info')
        self.request.response.redirect(context.absolute_url())



from ftw.upgrade import UpgradeStep
from izug.ticketbox.interfaces import ITicketboxDatagridIdUpdater


class RecalculateTicketboxDatagridIds(UpgradeStep):
    """Recalculate ticketbox datagrid ids.
    """

    def __call__(self):
        self.install_upgrade_profile()
        for obj in self.objects(
                {'portal_type': 'Ticket Box'},
                'Recalculate ticketbox datagrid ids'):
            ITicketboxDatagridIdUpdater(obj)()



from izug.ticketbox.interfaces import ITicket
from izug.ticketbox.interfaces import ITicketReferenceStartupDirectory
from zope.component import adapts
from zope.interface import Interface
from zope.interface import implements


class DefaultTicketReferenceStartupDirectory(object):
    implements(ITicketReferenceStartupDirectory)
    adapts(ITicket, Interface)

    def __init__(self, ticket, request):
        self.ticket = ticket
        self.request = request

    def get(self):
        url_tool = self.ticket.restrictedTraverse('@@plone_tools').url()
        path = '/'.join(url_tool.getRelativeContentPath(self.ticket))
        return path



from ftw.builder.archetypes import ArchetypesBuilder
from ftw.builder import builder_registry


class TicketBoxBuilder(ArchetypesBuilder):
    portal_type = 'Ticket Box'

builder_registry.register('ticket box', TicketBoxBuilder)


class TicketBuilder(ArchetypesBuilder):
    portal_type = 'Ticket'

builder_registry.register('ticket', TicketBuilder)



from ftw.builder import Builder
from ftw.builder import create
from izug.ticketbox.testing import TICKETBOX_FUNCTIONAL_TESTING
from izug.ticketbox.tests import helpers
from plone.app.testing import TEST_USER_NAME, TEST_USER_PASSWORD
from plone.testing.z2 import Browser
from StringIO import StringIO
from unittest2 import TestCase
import transaction


class TestAttachmentCreation(TestCase):

    layer = TICKETBOX_FUNCTIONAL_TESTING

    def setUp(self):
        super(TestAttachmentCreation, self).setUp()

        portal = self.layer['portal']
        helpers.login_as_manager(portal)

        self.ticketbox = helpers.create_ticketbox(portal)

        self.browser = Browser(self.layer['app'])
        self.browser.handleErrors = False
        self.browser.addHeader('Authorization', 'Basic %s:%s' % (
                TEST_USER_NAME, TEST_USER_PASSWORD,))

        transaction.commit()

    def test_attachment_creation(self):
        ticketbox = create(Builder('ticket box')
                           .titled('THe Ticket Box'))

        ticket = create(Builder('ticket')
                        .titled('The Ticket')
                        .within(ticketbox))

        self.browser.open(ticket.absolute_url() + '/folder_factories')
        self.browser.getControl('TicketAttachment').click()
        self.browser.getControl(name='form.button.Add').click()

        file_field = self.browser.getControl(name="file_file")
        file_field.add_file(
            StringIO('GIF89a\x01\x00\x01\x00\x80\x00\x00\x00\x00\x00\x00\x00'
                            '\x00!\xf9\x04\x04\x00\x00\x00\x00,\x00\x00\x00'
                            '\x00\x01\x00\x01\x00\x00\x02\x02\x01\x00;'),
            'image/gif',
            'peter.gif')
        self.browser.getControl(name="form.button.save").click()
        self.assertEqual(
            self.browser.url,
            'http://nohost/plone/the-ticket-box/the-ticket/peter.gif/view')



from ftw.notification.email.templates.base import BaseEmailRepresentation
from izug.ticketbox import ticketboxMessageFactory as _
from izug.ticketbox.browser.helper import map_attribute, readable_author
from zope.app.pagetemplate import ViewPageTemplateFile
from zope.i18nmessageid import Message
from Products.CMFCore.utils import getToolByName
from Acquisition import aq_inner


class TicketEmailRepresentation(BaseEmailRepresentation):

    template = ViewPageTemplateFile('ticket.pt')

    def getTracker(self):
        context = self.context
        tracker = context.aq_parent

        return tracker

    def get_responses(self):
        base_response = self.context.restrictedTraverse('base_response')
        return base_response.responses()

    def creator(self):
        if self.infos()['response'] is True:
            responses = self.get_responses()
            response = responses[len(responses) - 1]
            userid = response['response'].creator
        else:
            context = aq_inner(self.context)
            userid = context.Creator()

        mem_tool = getToolByName(self.context, 'portal_membership')
        member = mem_tool.getMemberById(userid)
        return member.getUser().getProperty('fullname', userid)

    def infos(self):
        """Returns Infos for email-template"""
        responses = self.get_responses()
        author = readable_author(self.context)

        if isinstance(author, Message):
            author = self.context.translate(author)

        ticket_infos = {'tracker_title': self.context.aq_parent.title,
                        'tracker_url': self.context.aq_parent.absolute_url(),
                        'title': self.context.Title(),
                        'ticket_id': self.context.getId(),
                        'individualIdendifier':
                            self.context.aq_parent.getIndividualIdentifier,
                        'url': self.context.absolute_url(),
                        'text': self.context.getTicket_description(),
                        'state': map_attribute(self.context, "state"),
                        'responsibleManager': author,
                        'priority': map_attribute(self.context, "priority"),
                        'area': map_attribute(self.context, "area"),
                        'variety': map_attribute(self.context, "variety"),
                        'releases': map_attribute(self.context, "releases"),
                        'watchedRelease': map_attribute(self.context,
                                                        "watchedRelease"),
                        'answerDate': self.context.toLocalizedTime(
                            self.context.getAnswerDate(), long_format=True),
                        'response': False}
        if responses == []:
            return ticket_infos
        else:
            response_date = responses[len(responses) - 1]['response'].date
            if self.context.modification_date > response_date:
                return ticket_infos
            else:
                response = responses[len(responses) - 1]
                changes = {
                    'tracker_title': self.context.aq_parent.title,
                    'tracker_url': self.context.aq_parent.absolute_url(),
                    'title': self.context.Title(),
                    'ticket_id': self.context.getId(),
                    'individualIdendifier':
                        self.context.aq_parent.getIndividualIdentifier,
                    'url': self.context.absolute_url(),
                    'text': '',
                    'state': '',
                    'responsibleManager': '',
                    'priority': '',
                    'area': '',
                    'variety': '',
                    'releases': '',
                    'watchedRelease': '',
                    'answerDate': '',
                    'response': True}
                for item in response['response'].changes:
                    # XXX: Hack to solve the label_unassigned translations
                    # problem.
                    # If we retrieve a responsibleManager named
                    # label_unassigned, try to translate it
                    if item['id'] == 'responsibleManager':
                        if item['before'] == 'label_unassigned':
                            item['before'] = self.context.translate(
                                _(item['before']))
                        if item['after'] == 'label_unassigned':
                            item['after'] = self.context.translate(
                                _(item['after']))

                    changes[item['id']] = (
                        item['before'] + ' &rarr; ' + item['after'])
                changes['text'] = response['response'].text
                return changes



from izug.ticketbox.testing import TICKETBOX_INTEGRATION_TESTING
from plone.app.testing import login, setRoles
from plone.app.testing import TEST_USER_NAME, TEST_USER_ID
from Products.CMFCore.utils import getToolByName
from unittest2 import TestCase
import os


class TestGetIcon(TestCase):

    layer = TICKETBOX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.file_ = open(
            "%s/testfile.NonsenseExtension" % os.path.split(__file__)[0], 'r'
            )
        self.mtr = getToolByName(self.portal, 'mimetypes_registry')
        ticketattachment_fti = getattr(self.portal.portal_types,
                                       'TicketAttachment')
        ticketattachment_fti.global_allow = True
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        login(self.portal, TEST_USER_NAME)

    def test_custom_icon(self):
        self.mtr.manage_addMimeType('customimage type',
                                    ['image/x-custom-image'],
                                    ['NonsenseExtension'],
                                    'customicon.png')
        plonefile = self.portal.get(
            self.portal.invokeFactory('TicketAttachment',
                                      'myattachment',
                                      file=self.file_)
            )
        self.assertEqual('plone/customicon.png', plonefile.getIcon())

    def test_default_icon(self):
        self.mtr.manage_addMimeType('customimage type',
                                    ['image/x-custom-image'],
                                    ['NonsenseExtension'],
                                    '')
        plonefile = self.portal.get(
            self.portal.invokeFactory('TicketAttachment',
                                      'myattachment',
                                      file=self.file_)
            )
        self.assertEqual('plone/image.png', plonefile.getIcon())

    def test_contenttype_not_found(self):
        #To test the behavior of our function when a contenttype is not found
        #in mimetyperegistry we add a mimetype so everything is indexed and
        #created correctly.
        self.mtr.manage_addMimeType('customimage type',
                                    ['image/x-custom-image'],
                                    ['NonsenseExtension'], '')
        plonefile = self.portal.get(
            self.portal.invokeFactory('TicketAttachment',
                                      'myattachment',
                                      file=self.file_)
            )
        #and then we remove it again so we won't find it in the registry.
        self.mtr.manage_delObjects(['image/x-custom-image'])

        self.assertIn(plonefile.getIcon(), (
                'plone/image_icon.gif',  # Plone <= 4.3.2
                'plone/image_icon.png',  # Plone >= 4.3.3
                ))



from copy import deepcopy
from ftw.builder import Builder
from ftw.builder import create
from izug.ticketbox.interfaces import ITicketboxDatagridIdUpdater
from izug.ticketbox.testing import TICKETBOX_FUNCTIONAL_TESTING
from izug.ticketbox.tests import helpers
from izug.ticketbox.tests.data import STATE_DATA
from Products.Archetypes.event import ObjectEditedEvent
from unittest2 import TestCase
from zope.event import notify


class TestTicketboxDatagridIdUpdater(TestCase):

    layer = TICKETBOX_FUNCTIONAL_TESTING

    def setUp(self):
        super(TestTicketboxDatagridIdUpdater, self).setUp()
        portal = self.layer['portal']
        helpers.login_as_manager(portal)

    def test_generate_ids_on_creation(self):
        states = [
            {'id': '',
             'title': "Chuck Norris",
             'show_in_all_tickets': '1',
             'show_in_my_tickets': '0',
             }]

        ticketbox = create(Builder('ticket box')
                           .titled('The Ticket Box')
                           .having(
                               availableStates=states))

        self.assertEqual(
            'chuck-norris',
            ticketbox.getAvailableStates()[0].get('id'))

    def test_regenerate_datagrid_ids_if_title_was_modified(self):
        states = [
            {'id': 'chuck-norris',
             'title': "Chuck Norris",
             'show_in_all_tickets': '1',
             'show_in_my_tickets': '0',
             }]

        ticketbox = create(Builder('ticket box')
                           .titled('The Ticket Box')
                           .having(
                               availableStates=states))

        ticketbox.getAvailableStates()[0]['title'] = "James Bond"
        notify(ObjectEditedEvent(ticketbox))

        self.assertEqual(
            'james-bond',
            ticketbox.getAvailableStates()[0].get('id'))

    def test_update_tickets_if_title_of_ticketbox_datagrid_was_modified(self):
        states = [
            {'id': 'chuck-norris',
             'title': "Chuck Norris",
             'show_in_all_tickets': '1',
             'show_in_my_tickets': '0',
             }]

        ticketbox = create(Builder('ticket box')
                           .titled('The Ticket Box')
                           .having(
                               availableStates=states))

        ticket = create(Builder('ticket')
                        .within(ticketbox)
                        .having(state='chuck-norris'))

        ticketbox.getAvailableStates()[0]['title'] = "James Bond"
        notify(ObjectEditedEvent(ticketbox))

        self.assertEqual('james-bond', ticket.getState())

    def test_do_not_update_tickets_from_other_ticketboxes(self):
        states = [
            {'id': 'chuck-norris',
             'title': "Chuck Norris",
             'show_in_all_tickets': '1',
             'show_in_my_tickets': '0',
             }]

        ticketbox_1 = create(Builder('ticket box')
                             .titled('The Ticket Box 1')
                             .having(
                                 availableStates=list(deepcopy(states))))

        ticketbox_2 = create(Builder('ticket box')
                             .titled('The Ticket Box 2')
                             .having(
                                 availableStates=list(deepcopy(states))))

        ticket_1 = create(Builder('ticket')
                          .within(ticketbox_1)
                          .having(state='chuck-norris'))

        ticket_2 = create(Builder('ticket')
                          .within(ticketbox_2)
                          .having(state='chuck-norris'))

        ticketbox_1.getAvailableStates()[0]['title'] = "James Bond"
        notify(ObjectEditedEvent(ticketbox_1))
        notify(ObjectEditedEvent(ticketbox_2))

        self.assertEqual('james-bond', ticket_1.getState())
        self.assertEqual('chuck-norris', ticket_2.getState())


class TestUpdate(TestCase):

    layer = TICKETBOX_FUNCTIONAL_TESTING

    def setUp(self):
        super(TestUpdate, self).setUp()
        portal = self.layer['portal']
        helpers.login_as_manager(portal)

        self.ticketbox = create(Builder('ticket box')
                                .titled('The Ticket Box')
                                .having(
                                    availableStates=list(deepcopy(STATE_DATA))))

        self.adapter = ITicketboxDatagridIdUpdater(self.ticketbox)

    def test_do_not_update_if_nothing_changed(self):
        ticket = create(Builder('ticket')
                        .within(self.ticketbox)
                        .having(state='show-all-tickets'))

        self.assertEqual(0, len(self.adapter.update()))
        self.assertEqual('show-all-tickets', ticket.getState())

    def test_update_all_tickets_if_something_changed_on_datagrid(self):
        ticket_1 = create(Builder('ticket')
                          .within(self.ticketbox)
                          .having(state='show-all-tickets'))

        ticket_2 = create(Builder('ticket')
                          .within(self.ticketbox)
                          .having(state='show-my-tickets'))

        self.ticketbox.getAvailableStates()[0]['title'] = "Chuck Norris"

        self.assertEqual(1, len(self.adapter.update()))
        self.assertEqual('chuck-norris', ticket_1.getState())
        self.assertEqual(
            'chuck-norris',
            self.ticketbox.getAvailableStates()[0].get('id'))


class TestUpdateTickets(TestCase):

    layer = TICKETBOX_FUNCTIONAL_TESTING

    def setUp(self):
        super(TestUpdateTickets, self).setUp()
        portal = self.layer['portal']
        helpers.login_as_manager(portal)

        self.ticketbox = create(Builder('ticket box')
                                .titled('THe Ticket Box'))

        self.adapter = ITicketboxDatagridIdUpdater(self.ticketbox)

    def test_do_not_update_tickets_if_nothing_changed(self):
        self.adapter.changes = {}

        self.assertEqual([], self.adapter.update_tickets())

    def test_do_not_update_ticket_if_no_related_attribute_changed(self):
        ticket = create(Builder('ticket')
                        .within(self.ticketbox)
                        .having(state='foo'))

        self.adapter.changes = {'availableAreas': [('foo', 'bar')]}

        self.assertEqual(0, len(self.adapter.update_tickets()))
        self.assertEqual('foo', ticket.getState())

    def test_do_not_update_ticket_if_the_id_was_never_generated_before(self):
        ticket = create(Builder('ticket')
                        .within(self.ticketbox)
                        .having(state='foo'))

        self.adapter.changes = {'availableStates': [('', 'bar')]}

        self.assertEqual(0, len(self.adapter.update_tickets()))
        self.assertEqual('foo', ticket.getState())

    def test_do_not_update_ticket_if_the_changed_id_is_not_related_to_the_ticket(self):
        ticket = create(Builder('ticket')
                        .within(self.ticketbox)
                        .having(state='foo'))

        self.adapter.changes = {'availableStates': [('bar', 'chuck')]}

        self.assertEqual(0, len(self.adapter.update_tickets()))
        self.assertEqual('foo', ticket.getState())

    def test_update_ticket_if_related_id_changed(self):
        ticket = create(Builder('ticket')
                        .within(self.ticketbox)
                        .having(state='foo'))

        self.assertEqual(1, len(self.adapter.update_tickets(
            {'availableStates': [('foo', 'bar')]})))
        self.assertEqual('bar', ticket.getState())

    def test_update_multiple_tickets_works_properly(self):
        ticket_1 = create(Builder('ticket')
                          .within(self.ticketbox)
                          .having(state='foo'))

        ticket_2 = create(Builder('ticket')
                          .within(self.ticketbox)
                          .having(state='foo'))

        ticket_3 = create(Builder('ticket')
                          .within(self.ticketbox)
                          .having(state='chuck'))

        self.assertEqual(2, len(self.adapter.update_tickets(
            {'availableStates': [('foo', 'bar')]})))
        self.assertEqual('bar', ticket_1.getState())
        self.assertEqual('bar', ticket_2.getState())
        self.assertEqual('chuck', ticket_3.getState())


class TestUniquifyIds(TestCase):

    layer = TICKETBOX_FUNCTIONAL_TESTING

    def setUp(self):
        super(TestUniquifyIds, self).setUp()
        portal = self.layer['portal']
        helpers.login_as_manager(portal)

        ticketbox = create(Builder('ticket box')
                           .titled('THe Ticket Box'))

        self.adapter = ITicketboxDatagridIdUpdater(ticketbox)

    def test_new_ids_set(self):
        input = [{'id': '', 'title': 'Foo'},
                 {'id': '', 'title': 'Bar'}]

        updated = [{'id': 'foo', 'title': 'Foo'},
                   {'id': 'bar', 'title': 'Bar'}]

        self.adapter.uniquify_ids(input)

        self.assertEqual(updated, input)

    def test_partical_ids(self):
        input = [{'id': 'foo', 'title': 'Foo'},
                 {'id': '', 'title': 'Bar'}]

        updated = [{'id': 'foo', 'title': 'Foo'},
                   {'id': 'bar', 'title': 'Bar'}]

        self.adapter.uniquify_ids(input)

        self.assertEqual(updated, input)

    def test_conflicting_ids(self):
        input = [{'id': '', 'title': 'Foo'},
                 {'id': '', 'title': 'Foo'},
                 {'id': '', 'title': 'Foo'}]

        updated = [{'id': 'foo', 'title': 'Foo'},
                   {'id': 'foo-1', 'title': 'Foo'},
                   {'id': 'foo-2', 'title': 'Foo'}]

        self.adapter.uniquify_ids(input)

        self.assertEqual(updated, input)

    def test_modification_updateds_ids(self):
        input = [{'id': 'foo', 'title': 'Bar'}]

        updated = [{'id': 'bar', 'title': 'Bar'}]

        self.adapter.uniquify_ids(input)

        self.assertEqual(updated, input)

    def test_keeps_other_keys(self):
        input = [{'id': 'foo', 'title': 'Foo', 'bar': 'Bar'}]
        updated = [{'id': 'foo', 'title': 'Foo', 'bar': 'Bar'}]

        self.adapter.uniquify_ids(input)

        self.assertEqual(updated, input)

    def test_prevent_conflicting_if_rename_to_existing(self):
        input = [{'id': 'foo', 'title': 'Bar'},
                 {'id': 'bar', 'title': 'Foo'}]

        updated = [{'id': 'bar-1', 'title': 'Bar'},
                   {'id': 'foo-1', 'title': 'Foo'}]

        self.adapter.uniquify_ids(input)

        self.assertEqual(updated, input)

    def test_does_not_copy(self):
        input = [{'id': '', 'title': 'Foo'}]

        id_input = id(input)

        self.adapter.uniquify_ids(input)

        self.assertEqual(id_input, id(input))

    def test_return_value_is_empty_if_nothing_has_changed(self):
        input = [{'id': 'foo', 'title': 'Foo'}]

        changes = self.adapter.uniquify_ids(input)

        self.assertEqual([], changes)

    def test_return_value_contains_all_changes(self):
        input = [{'id': 'foo', 'title': 'Foo'},
                 {'id': 'bar-1', 'title': 'Bar'},
                 {'id': '', 'title': 'Chuck'}]

        changes = self.adapter.uniquify_ids(input)

        self.assertEqual([('bar-1', 'bar'), ('', 'chuck')], changes)


class TestCreateUniqueId(TestCase):

    layer = TICKETBOX_FUNCTIONAL_TESTING

    def setUp(self):
        super(TestCreateUniqueId, self).setUp()
        portal = self.layer['portal']
        helpers.login_as_manager(portal)

        ticketbox = create(Builder('ticket box')
                           .titled('THe Ticket Box'))

        self.adapter = ITicketboxDatagridIdUpdater(ticketbox)

    def test_first_id(self):
        self.assertEqual(self.adapter.create_uniqe_id('Foo', []), 'foo')

    def test_existing_id(self):
        self.assertEqual(
            self.adapter.create_uniqe_id('Foo', ['foo']), 'foo-1')

        self.assertEqual(
            self.adapter.create_uniqe_id('Foo', ['foo', 'foo-1']), 'foo-2')

    def test_spaces(self):
        self.assertEqual(
            self.adapter.create_uniqe_id('Foo bar', []), 'foo-bar')



from ftw.builder import Builder
from ftw.builder import create
from izug.ticketbox.testing import TICKETBOX_INTEGRATION_TESTING
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import setRoles
from unittest2 import TestCase



class TestReferenceStartupDirectory(TestCase):

    layer = TICKETBOX_INTEGRATION_TESTING

    def setUp(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        login(portal, TEST_USER_NAME)

    def test_startup_directory_is_current_context(self):
        ticketbox = create(Builder('ticket box')
                           .titled('THe Ticket Box'))

        ticket = create(Builder('ticket')
                        .titled('The Ticket')
                        .within(ticketbox))

        self.assertEquals('the-ticket-box/the-ticket',
                          ticket.getReferenceStartupDirectory())



from ftw.notification.base.browser.views import NotificationForm as base


class NotificationForm(base):
    """Notification with preselect for tickets"""

    def __init__(self, context, request):
        super(NotificationForm, self).__init__(context, request)
        self.pre_select = []

        self.pre_select.append(self.context.getResponsibleManager())




from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import applyProfile
from plone.testing import z2
from zope.configuration import xmlconfig
import izug.ticketbox.tests.builder


class TicketBoxLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE, BUILDER_LAYER)

    def setUpZope(self, app, configurationContext):
        xmlconfig.string(
            '<configure xmlns="http://namespaces.zope.org/zope">'
            '  <include package="z3c.autoinclude" file="meta.zcml" />'
            '  <includePlugins package="plone" />'
            '  <includePluginsOverrides package="plone" />'

            '  <include package="Products.DataGridField" />'
            '</configure>',
            context=configurationContext)

        z2.installProduct(app, 'ftw.notification.base')
        z2.installProduct(app, 'ftw.notification.email')
        z2.installProduct(app, 'ftw.calendarwidget')
        z2.installProduct(app, 'izug.ticketbox')
        z2.installProduct(app, 'collective.MockMailHost')
        z2.installProduct(app, 'ftw.zipexport')

    def setUpPloneSite(self, portal):
        # Install into Plone site using portal_setup
        applyProfile(portal, 'ftw.notification.base:default')
        applyProfile(portal, 'ftw.notification.email:default')
        applyProfile(portal, 'izug.ticketbox:default')
        applyProfile(portal, 'collective.MockMailHost:default')
        applyProfile(portal, 'ftw.zipexport:default')

TICKETBOX_FIXTURE = TicketBoxLayer()
TICKETBOX_INTEGRATION_TESTING = IntegrationTesting(
    bases=(TICKETBOX_FIXTURE, ), name="TicketBox:Integration")

TICKETBOX_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(TICKETBOX_FIXTURE,
           set_builder_session_factory(functional_session_factory)),
    name="TicketBox:Functional")



from AccessControl import getSecurityManager
from Acquisition import aq_inner, aq_parent
from copy import deepcopy
from DateTime import DateTime
from izug.ticketbox.interfaces import ITicket, IResponseContainer, IResponse
from izug.ticketbox.interfaces import ITicketboxDatagridIdUpdater
from persistent import Persistent
from persistent.list import PersistentList
from plone.i18n.normalizer.interfaces import IIDNormalizer
from Products.CMFCore.utils import getToolByName
from zope.annotation.interfaces import IAnnotations
from zope.app.component import hooks
from zope.app.component.hooks import getSite
from zope.app.container.contained import ObjectAddedEvent
from zope.app.container.contained import ObjectRemovedEvent
from zope.app.container.interfaces import UnaddableError
from zope.component import adapts
from zope.component import queryUtility
from zope.event import notify
from zope.interface import implements


class TicketboxDatagridIdUpdater(object):
    """A Ticketbox has a lot of datagrid fields to define different
    attributes dynamically (i.e. state, priorities...).

    For each datagrid-row we generate an id wich will be used
    to query the catalog and to reference in a ticket.

    If the user changes a title of a ticketbox datagrid row, we have
    to change the id too. Otherwise, orderings and groupings wont
    work properly again.

    We look for changes in the datagrid fields and updates the ids.
    If we have found changes, we update all tickets which where referenced
    with one of the changed ids.

    """

    implements(ITicketboxDatagridIdUpdater)

    ticketbox_ticket_field_mapping = {
        'availableStates': 'state',
        'availableReleases': 'releases',
        'availablePriorities': 'priority',
        'availableAreas': 'area',
        'availableVarieties': 'variety',
    }

    def __init__(self, context):
        self.context = context
        self.catalog = getToolByName(context, 'portal_catalog')

    def __call__(self):
        self.update()

    def update(self):
        """Search for changed rows in the ticketbox and updates the datagrid-ids
        and all its related tickets
        """

        # changed_fields contains all changed ids in the datagrid.
        #
        # The key is the ticketbox field name which contains changes.
        # The value is a list with tuples containing the old id as
        # the first entry and the new id as the new id
        #
        # ``changed_fields`` example:
        #
        # >>> {'availableAreas': [('my-Area', 'my-cool-area')],
        # ...  'availablePriorities': [('', 'high'), ('lower', 'low')],
        # ...  'availableStates': [('a-state', 'd-state')]}
        changed_fields = {}
        for field_name in self.ticketbox_ticket_field_mapping.keys():
            values = self.context.getField(field_name).get(self.context)
            changed_rows = self.uniquify_ids(values)
            if changed_rows:
                changed_fields[field_name] = changed_rows

        return self.update_tickets(changed_fields)

    def update_tickets(self, changed_fields={}):
        """Updates all ticket ids referencing to an old ticketbox
        datagrid id and reindexes modified tickets.
        """
        updated_tickets = []
        if not changed_fields:
            # Do nothing if nothing changed in the datagrid
            return updated_tickets

        for ticket in self.context.listFolderContents({'portal_type': 'Ticket'}):
            updated_fields = []

            # Iterate over all changed attributes
            for field_name, changes in changed_fields.items():

                # Do nothing if there where no changes for this attribute
                if not changes:
                    continue

                # Iterate over all changes for this attribute
                for old_id, new_id in changes:
                    if not old_id:
                        # This seems to be a new entry and there was no
                        # id generated yet. We don't have to do anything
                        continue

                    field = ticket.getField(
                        self.ticketbox_ticket_field_mapping.get(field_name))

                    current_id = field.get(ticket)

                    if old_id != current_id:
                        # Do not update the ticket if it is not related with
                        # a changed datagrid row.
                        continue

                    # Set the new id to the tickets field
                    field.set(ticket, new_id)

                    updated_fields.append(field)

            if updated_fields:
                updated_tickets.append(ticket)
                self.reindex_ticket(ticket, updated_fields)

        return updated_tickets

    def reindex_ticket(self, ticket, updated_fields):
        idxs = []
        for field in updated_fields:
            idxs.append(field.accessor)

        self.catalog.reindexObject(ticket, idxs=idxs)

    def uniquify_ids(self, data):
        """Creates unique ids within a list of dicts.
        Sets the "id" key of each dicts, creates the id with the value of
        "title" within the dict.

        ``data`` example:

        >>> [{'id': '', 'title': 'Foo'},
        ...  {'id': 'bar', 'title': 'Bar'}]

        if the ID is already set but the title changed, it will update
        the id with a newly generated one.

        example:

        >>> uniquify_ids([
        ...    {'id': 'foo', 'title': 'Chuck'},
               {'id': 'bar', 'title': 'Bar'}])

        >>> [{'id': 'chuck', 'title': 'Chuck'}, [{'id': 'bar', 'title': 'Bar'},

        if the title changed to an already existing title it will update to a
        new unique id:

        >>> uniquify_ids([
        ...    {'id': 'foo', 'title': 'Bar'},
               {'id': 'bar', 'title': 'Bar'}])

        >>> [{'id': 'bar-1', 'title': 'Bar'}, [{'id': 'bar', 'title': 'Bar'},


        if the title didn't change, it will do nothing.
        """
        existing_ids = set([item.get('id') for item in data if item.get('id')])
        changes = []
        for item in data:
            old_id = item.get('id')

            existing_ids_without_old_id = deepcopy(existing_ids)
            existing_ids_without_old_id.difference_update([old_id])

            new_id = self.create_uniqe_id(
                item.get('title'), existing_ids_without_old_id)

            if old_id == new_id:
                continue

            item['id'] = new_id
            existing_ids.add(new_id)

            changes.append((old_id, new_id))

        return changes

    def create_uniqe_id(self, title, existing_ids):
        """Creates a uniqe id by using the IIDNormalizer utility.
        """

        id_ = queryUtility(IIDNormalizer).normalize(title)

        if id_ not in existing_ids:
            return id_

        base_id = id_
        index = 0

        while id_ in existing_ids:
            index += 1
            id_ = '%s-%i' % (base_id, index)

        return id_


class ResponseContainer(Persistent):

    implements(IResponseContainer)
    adapts(ITicket)
    ANNO_KEY = 'izug.ticketbox.responses'

    def __init__(self, context):
        self.context = context
        annotations = IAnnotations(self.context)
        self.__mapping = annotations.get(self.ANNO_KEY, None)
        if self.__mapping is None:
            self.__mapping = PersistentList()
            annotations[self.ANNO_KEY] = self.__mapping

    def __contains__(self, key):
        '''See interface IReadContainer

        Taken from zope.app.container.btree.

        Reimplement this method, since has_key() returns the key if available,
        while we expect True or False.

        >>> c = ResponseContainer()
        >>> "a" in c
        False
        >>> c["a"] = 1
        >>> "a" in c
        True
        >>> "A" in c
        False
        '''
        return key in self.__mapping

    has_key = __contains__

    def __getitem__(self, i):
        i = int(i)
        return self.__mapping.__getitem__(i)

    def __delitem__(self, item):
        self.__mapping.__delitem__(item)

    def __len__(self):
        return self.__mapping.__len__()

    def __setitem__(self, i, y):
        self.__mapping.__setitem__(i, y)

    def append(self, item):
        self.__mapping.append(item)

    def remove(self, id_):
        """Remove item 'id_' from the list.

        We don't actually remove the item, we just set it to None,
        so that when you edit item 3 out of 3 and someone deletes
        item 2 you are not left in the water.

        Note that we used to get passed a complete item, not an id.
        """
        id_ = int(id_)
        self[id_] = None

    def add(self, item):
        if not IResponse.providedBy(item):
            raise UnaddableError(self, item,
                                 "IResponse interface not provided.")
        self.append(item)
        id_ = str(len(self))
        event = ObjectAddedEvent(item, newParent=self.context, newName=id_)
        notify(event)

    def delete(self, id_):
        # We need to fire an ObjectRemovedEvent ourselves here because
        # self[id_].__parent__ is not exactly the same as self, which
        # in the end means that __delitem__ does not fire an
        # ObjectRemovedEvent for us.
        #
        # Also, now we can say the oldParent is the issue instead of
        # this adapter.
        event = ObjectRemovedEvent(self[id_], oldParent=self.context,
                                   oldName=id_)
        self.remove(id_)
        notify(event)


class Response(Persistent):

    implements(IResponse)

    def __init__(self, text):
        self.__parent__ = self.__name__ = None
        self.text = text
        self.changes = PersistentList()
        sm = getSecurityManager()
        user = sm.getUser()
        self.creator = user.getId() or '(anonymous)'
        self.date = DateTime()
        self.type = 'additional'
        self.mimetype = ''
        self.rendered_text = None
        self.attachment = None
        self.references = []

    def add_change(self, id_, name, before, after):
        """Add a new issue change.
        """
        delta = dict(
            id=id_,
            name=name,
            before=before,
            after=after)
        self.changes.append(delta)

    def creator_fullname(self):
        site = getSite()
        member = site.portal_membership.getMemberById(self.creator)
        # In case of Creator no longer exists
        if member is None:
            return self.creator
        fullname = member.getProperty('fullname', member.getId())
        if fullname:
            return fullname
        return member.getId()


class EmptyExporter(object):

    def __init__(self, context):
        self.context = context

    def export(self, export_context, subdir, root=False):
        return


class TicketBoxSubjectCreator(object):

    def __init__(self, context):
        self.context = aq_inner(context)
        self.request = self.context.REQUEST

    def __call__(self, object_):
        site = hooks.getSite()
        portal_properties = getToolByName(object_, 'portal_properties')
        default_subject = '[%s] Notification:' % (site.Title())
        subject = None
        try:
            sheet = portal_properties['ftw.notification-properties']

        except AttributeError:
            subject = default_subject

        else:
            individual_identifier = object_.getIndividualIdentifier()
            if individual_identifier:
                individual_identifier = ' [%s]' % individual_identifier
            subject = '%s%s %s' % (
                sheet.getProperty('notification_email_subject',
                                  default_subject),
                individual_identifier,
                object_.Title())

        return subject


class TicketSubjectCreator(object):

    def __init__(self, context):
        self.context = aq_inner(context)
        self.request = self.context.REQUEST

    def __call__(self, object_):
        site = hooks.getSite()
        portal_properties = getToolByName(object_, 'portal_properties')
        default_subject = '[%s] Notification:' % (site.Title())
        subject = None
        try:
            sheet = portal_properties['ftw.notification-properties']
        except AttributeError:
            subject = default_subject

        else:
            individual_identifier = aq_parent(
                self.context).getIndividualIdentifier()
            if individual_identifier:
                individual_identifier = ' [%s]' % individual_identifier

            subject = '%s%s #%s - %s' % (
                sheet.getProperty('notification_email_subject',
                                  default_subject),
                individual_identifier,
                object_.getId(),
                object_.Title())

        return subject



from ftw.upgrade import UpgradeStep


class FixIconsForTicketAndTicketBox(UpgradeStep):
    """Fix icons for ticket and ticket box.
    """

    def __call__(self):
        self.install_upgrade_profile()



from Acquisition import aq_parent
from zope.interface import implements
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleVocabulary


class TicketBoxVocabulary(object):

    implements(IVocabularyFactory)

    def __init__(self, field):
        self.field = field

    def __call__(self, context):
        tracker = aq_parent(context)
        myField = tracker.getField(self.field)
        result = []
        for item in myField.get(tracker):
            result.append(SimpleVocabulary.createTerm(
                    item['id'], item['id'], item['title']))
        if result == []:
            result.append(SimpleVocabulary.createTerm(
                    '', '', '-'))
        return result


StatesVocabularyFactory = TicketBoxVocabulary('availableStates')
ReleasesVocabularyFactory = TicketBoxVocabulary('availableReleases')
PrioritiesVocabularyFactory = TicketBoxVocabulary('availablePriorities')
AreasVocabularyFactory = TicketBoxVocabulary('availableAreas')
VarietiesVocabularyFactory = TicketBoxVocabulary('availableVarieties')



from ftw.zipexport.interfaces import IZipRepresentation
from ftw.zipexport.representations.archetypes import FolderZipRepresentation
from ftw.zipexport.representations.general import NullZipRepresentation
from izug.ticketbox.export import get_header, get_data, create_xlsx
from izug.ticketbox.interfaces import ITicketBox, ITicket
from Products.CMFPlone.utils import safe_unicode
from StringIO import StringIO
from zope.component import adapts
from zope.interface import implements, Interface


class TicketBoxZipRepresentation(FolderZipRepresentation):
    implements(IZipRepresentation)
    adapts(ITicketBox, Interface)

    def get_files(self, path_prefix=u"", recursive=True, toplevel=True):
        """
        Returns a list of two-value-tuples having the following values:
        - a relative path under which the file should show up in the zip
        - the data as either a file or a stream
        """
        header = get_header(self.context)
        data = get_data(self.context)
        xlsx = create_xlsx(header, data)

        filename = '{0}.xlsx'.format(self.context.getId())

        yield (u'{0}/{1}'.format(safe_unicode(path_prefix),
                                 safe_unicode(filename)), xlsx)

        # Recursively export folder contents.
        folder_contents = super(TicketBoxZipRepresentation, self).get_files(
            path_prefix, recursive, toplevel)

        for item in folder_contents:
            yield item


class TicketZipRepresentation(NullZipRepresentation):
    implements(IZipRepresentation)
    adapts(ITicket, Interface)

    def get_files(self, path_prefix=u"", recursive=False, toplevel=True):
        attachments = self.context.getAttachments()

        # Put the attachments in a folder per ticket.
        path_prefix = u'{0}/{1}'.format(safe_unicode(path_prefix),
                                        safe_unicode(self.context.Title()))

        for attachment in attachments:
            filename = attachment.getFilename()
            yield (u'{0}/{1}'.format(safe_unicode(path_prefix),
                                     safe_unicode(filename)),
                   StringIO(attachment.data))



from browser.helper import map_base
from izug.ticketbox import ticketboxMessageFactory as _
from Products.CMFCore.utils import getToolByName
import StringIO
import xlsxwriter
from zope.i18n import translate


def format_date(date):
    if not date:
        return ''
    if date.year() <= 1900:
        return ''
    else:
        return date.strftime('%d.%m.%Y %H:%M')


def fullname(context, uid):
    member = context.portal_membership.getMemberById(uid)
    if member:
        fullname = member.getProperty('fullname', member.getId())
        if fullname:
            return fullname
        return member.getId()
    return uid


def get_header(context):
    return [
        context.translate(
            _(u'export_heading_nr', default=u'No.')),
        context.translate(
            _(u'export_heading_title', default=u'Title')),
        context.translate(
            _(u'export_heading_description', default=u'Description')),
        context.translate(
            _(u'export_heading_creator', default=u'Creator')),
        context.translate(
            _(u'export_heading_created', default=u'Created')),
        context.translate(
            _(u'export_heading_responsible', default=u'Responsible')),
        context.translate(
            _(u'export_heading_state', default=u'State')),
        context.translate(
            _(u'export_heading_priorities', default=u'Priorities')),
        context.translate(
            _(u'export_heading_area', default=u'Area')),
        context.translate(
            _(u'export_heading_variety', default=u'Variety')),
        context.translate(
            _(u'export_heading_releases', default=u'Releases')),
        context.translate(
            _(u'export_heading_watched_release', default=u'Watched release')),
        context.translate(
            _(u'export_heading_watched_due_date', default=u'Due date')),
    ]


def get_data(context):
    data = []
    brains = list(context.portal_catalog(
        {
            'portal_type': 'Ticket',
            'sort_on': 'getId',
            'path': {
                'query': '/'.join(context.getPhysicalPath()),
                'depth': 2,
            }
        }
    ))

    pt = getToolByName(context, 'portal_transforms')
    for brain in brains:
        obj = brain.getObject()
        ticket_description = obj.getTicket_description()
        row = [
            brain.getId,
            brain.Title,
            pt.convertTo('text/plain', ticket_description).getData(),
            fullname(context, brain.Creator),
            format_date(brain.created),
            fullname(context, brain.getResponsibleManager),
            map_base(
                context.getAvailableStates(),
                brain.getState),
            map_base(
                context.getAvailablePriorities(),
                brain.getPriority),
            map_base(
                context.getAvailableAreas(),
                brain.getArea),
            map_base(
                context.getAvailableVarieties(),
                brain.getVariety),
            map_base(
                context.getAvailableReleases(),
                brain.getReleases),
            map_base(
                context.getAvailableReleases(),
                brain.getWatchedRelease),
            format_date(brain.getDueDate)
        ]
        data.append(row)

    return data


def create_xlsx(header=None, data=None):
    if not header:
        header = []
    if not data:
        data = []

    # Create an in-memory output file for the new workbook.
    output = StringIO.StringIO()

    # Even though the final file will be in memory the module uses temp
    # files during assembly for efficiency. To avoid this on servers that
    # don't allow temp files, for example the Google APP Engine, set the
    # 'in_memory' constructor option to True.
    workbook = xlsxwriter.Workbook(output, {'in_memory': False})
    worksheet = workbook.add_worksheet()

    # Add a bold format for the header.
    bold = workbook.add_format({'bold': True})

    # Write the header.
    for index, item in enumerate(header):
        worksheet.write(0, index, item, bold)

    # Write the data (which is a nested list).
    for row_index, row_item in enumerate(data):
        for column_index, column_item in enumerate(row_item):
            worksheet.write_string(
                row_index+1, column_index, column_item.decode('utf-8'))

    # Set the auto filter only if there is something to filter.
    if header and data:
        worksheet.autofilter(0, 0, len(data)+1, len(data[0])-1)

    workbook.close()

    # Rewind the buffer.
    output.seek(0)

    return output



from ftw.tabbedview.browser.listing import CatalogListingView
from ftw.table import helper
from izug.ticketbox import ticketboxMessageFactory as _
from izug.ticketbox.browser.tabbed.base import BaseTicketListingTab


def icon(item, value):
    img = u'<img src="%s/%s"/>' % (item.portal_url(), item.getIcon)
    link = u'<a href="%s/at_download">%s</a>' % (item.getURL(), img)
    return link


def linked_attachment(item, title):
    return '<a href="%s/at_download/file">%s</a>' % (
        item.getURL(), title)


def attachment_ticketnr(item, nothing):
    url = item.getURL()
    split_url = url.split("/")

    if len(split_url) >= 2:
        ticket_url = '/'.join(split_url[:-1])
        return '<a href="%s"># %s</a>' % (ticket_url, split_url[-2])

    else:
        return id


class AllTicketsTab(BaseTicketListingTab):
    """Tab listing all tickets in this ticketbox.
    """

    def get_base_query(self):
        query = super(AllTicketsTab, self).get_base_query()

        query['getState'] = [state['id']
                             for state in self.context.getAvailableStates()
                             if state['show_in_all_tickets'] == '1']

        return query


class MyTicketsTab(BaseTicketListingTab):
    """Tab listing all tickets where the current user is responsible.
    """

    def get_base_query(self):
        query = super(MyTicketsTab, self).get_base_query()

        member = self.context.restrictedTraverse(
            '@@plone_portal_state').member()
        query['getResponsibleManager'] = member.getId()
        query['getState'] = [state['id']
                             for state in self.context.getAvailableStates()
                             if state['show_in_my_tickets'] == '1']

        return query


class MyIssuedTicketsTab(BaseTicketListingTab):
    """Tab listing all tickets where the current user is the creator.
    """

    def get_base_query(self):
        query = super(MyIssuedTicketsTab, self).get_base_query()

        member = self.context.restrictedTraverse(
            '@@plone_portal_state').member()
        query['Creator'] = member.getId()
        query['getState'] = [state['id']
                             for state in self.context.getAvailableStates()
                             if state['show_in_my_tickets'] == '1']

        return query


class AttachmentsTab(CatalogListingView):

    types = ['TicketAttachment']

    show_selects = False
    show_menu = False

    sort_on = 'sortable_title'

    columns = (
        {'column': 'Type',
         'column_title': _(u"Type"),
         'sort_index': 'getContentType',
         'transform': icon,
         },

        {'column': 'Title',
         'column_title': _(u"Title"),
         'sort_index': 'sortable_title',
         'transform': linked_attachment,
         },

        # There is no index ticket-number. Use the helper for getting it.
        {'column': 'ticket-number',
         'column_title': _(u"Ticketnr"),
         'sort_index': 'sortable_ticket_references',
         'transform': attachment_ticketnr,
         })


class TicketboxesTab(CatalogListingView):

    types = ['Ticket Box']
    show_selects = False
    show_menu = False
    sort_on = 'sortable_title'

    columns = (
        {'column': 'Title',
         'column_title': _(u"Title"),
         'sort_index': 'sortable_title',
         'transform': linked_attachment},

        {'column': 'modified',
         'column_title': _(u"ModificationDate"),
         'transform': helper.readable_date},

        {'column': 'get_owner_index',
         'column_title': _(u"Responsible"),
         'transform': helper.readable_author},
        )



from DateTime import DateTime
from izug.ticketbox.testing import TICKETBOX_INTEGRATION_TESTING
from izug.ticketbox.tests import helpers
from unittest2 import TestCase


class TestTicket(TestCase):

    layer = TICKETBOX_INTEGRATION_TESTING

    def setUp(self):
        super(TestTicket, self).setUp()

        portal = self.layer['portal']

        helpers.login_as_manager(portal)
        self.ticketbox = helpers.create_ticketbox(portal)
        self.ticket = helpers.create_ticket(self.ticketbox)

    def tearDown(self):
        helpers.remove_obj(self.ticketbox)
        helpers.logout_manager(self.layer['portal'])
        super(TestTicket, self).tearDown()

    def test_required_fields(self):
        title = self.ticket.getField('title')
        description = self.ticket.getField('description')

        self.assertEquals(title.required, True)
        self.assertEquals(description.required, False)

    def test_dates(self):
        due_date = self.ticket.getField('dueDate').get(self.ticket)
        answer_date = self.ticket.getField('answerDate').get(self.ticket)
        date = DateTime() + 14

        self.assertEquals(due_date.Date(), date.Date())
        self.assertEquals(answer_date.Date(), date.Date())



from izug.ticketbox.testing import TICKETBOX_FUNCTIONAL_TESTING
from izug.ticketbox.tests import helpers
from plone.app.testing import TEST_USER_NAME, TEST_USER_PASSWORD
from plone.testing.z2 import Browser
from unittest2 import TestCase
from zope.annotation.interfaces import IAnnotations
import transaction


class TestFunctionalTickets(TestCase):

    layer = TICKETBOX_FUNCTIONAL_TESTING

    def setUp(self):
        super(TestFunctionalTickets, self).setUp()

        portal = self.layer['portal']
        helpers.login_as_manager(portal)

        self.ticketbox = helpers.create_ticketbox(portal)

        self.browser = Browser(self.layer['app'])
        self.browser.handleErrors = False
        self.browser.addHeader('Authorization', 'Basic %s:%s' % (
                TEST_USER_NAME, TEST_USER_PASSWORD,))

        transaction.commit()

    def tearDown(self):
        helpers.remove_obj(self.ticketbox)
        helpers.logout_manager(self.layer['portal'])
        super(TestFunctionalTickets, self).tearDown()

    def test_view(self):
        ticketbox_url = self.ticketbox.absolute_url()

        self.browser.open(ticketbox_url + '/folder_factories')
        self.browser.getControl('Ticket').click()
        self.browser.getControl('Add').click()

        self.browser.getControl('Title').value = "Testissue"
        self.browser.getControl('Save').click()
        self.assertEqual('Submitted at' in self.browser.contents, True)

        self.browser.getControl(name="state").value = [
            self.browser.getControl(name="state").options[1]]
        self.browser.getControl(name="priority").value = [
            self.browser.getControl(name="priority").options[1]]
        self.browser.getControl(name="submit").click()
        self.assertIn('<td>Show My Tickets</td>',
                         self.browser.contents)
        self.assertIn('<td>Low</td>', self.browser.contents)

        ticket = self.ticketbox['1']
        annotations = IAnnotations(ticket)
        response = annotations['izug.ticketbox.responses'][0]

        self.assertEqual(response.changes[0]['before'], u'High')
        self.assertEqual(response.changes[1]['before'], u'Show All Tickets')
        self.assertEqual(response.changes[0]['after'], u'Low')
        self.assertEqual(response.changes[1]['after'], u'Show My Tickets')



from DateTime import DateTime
from ftw.builder import Builder
from ftw.builder import create
from ftw.testbrowser import browsing
from ftw.testbrowser.pages import statusmessages
from izug.ticketbox.adapters import Response
from izug.ticketbox.interfaces import IResponseContainer
from izug.ticketbox.testing import TICKETBOX_FUNCTIONAL_TESTING
from izug.ticketbox.tests import helpers
from unittest2 import TestCase


class TestResponse(TestCase):

    layer = TICKETBOX_FUNCTIONAL_TESTING

    def setUp(self):
        super(TestResponse, self).setUp()

        portal = self.layer['portal']
        helpers.login_as_manager(portal)

        self.ticketbox = helpers.create_ticketbox(portal)
        self.ticket1 = helpers.create_ticket(self.ticketbox)
        self.ticket2 = helpers.create_ticket(self.ticketbox, data_index=1)

        self.container1 = IResponseContainer(self.ticket1)
        self.container2 = IResponseContainer(self.ticket2)

    def tearDown(self):
        helpers.remove_obj(self.ticketbox)
        helpers.logout_manager(self.layer['portal'])
        super(TestResponse, self).tearDown()

    def test_add_and_remove(self):
        response1 = Response('response1')
        response2 = Response('response3')
        response3 = Response('response3')

        self.assertEquals(len(self.container1), 0)
        self.assertEquals(len(self.container2), 0)

        self.container1.add(response1)
        self.container1.add(response2)
        self.container2.add(response3)

        self.assertEquals(len(self.container1), 2)
        self.assertEquals(len(self.container2), 1)

        self.container1.remove(0)
        self.container1.remove(1)

        self.assertEquals(response1 in self.container1, False)
        self.assertEquals(response2 in self.container1, False)
        self.assertEquals(response3 in self.container2, True)

    @browsing
    def test_update_anserdate_if_given(self, browser):
        ticketbox = create(Builder('ticket box'))
        ticket = create(Builder('ticket')
                        .having(answerDate=DateTime(2015, 12, 31))
                        .within(ticketbox))

        browser.login().visit(ticket)

        browser.fill({
            'answerdate_year': '2021',
            'answerdate_month': 'April',
            'answerdate_day': '12',
            'answerdate_hour': '10',
            'answerdate_minute': '05'}).submit()

        self.assertEqual(
            1, len(browser.css('.response-reply')),
            "There should be one response visible")

        self.assertEqual(
            ['31.12.2015 00:00', '12.04.2021 10:05'],
            browser.css('.response-reply .issueChange').text,
            "The answerdate before and the answerdatedate after should "
            "be visible in the answer")

    @browsing
    def test_do_not_update_anserdate_if_no_new_date_is_given(self, browser):
        ticketbox = create(Builder('ticket box'))
        ticket = create(Builder('ticket')
                        .having(answerDate=DateTime(2015, 12, 31, 9, 50))
                        .within(ticketbox))

        browser.login().visit(ticket)

        browser.fill({
            'answerdate_year': '2015',
            'answerdate_month': 'December',
            'answerdate_day': '31',
            'answerdate_hour': '09',
            'answerdate_minute': '50'}).submit()

        self.assertEqual(
            ['No response text added and no issue changes made.'],
            statusmessages.error_messages())

    @browsing
    def test_do_not_update_anserdate_if_no_year_is_given(self, browser):
        ticketbox = create(Builder('ticket box'))
        ticket = create(Builder('ticket')
                        .having(answerDate=DateTime(2015, 12, 31))
                        .within(ticketbox))

        browser.login().visit(ticket)

        browser.fill({
            'answerdate_year': '--',
            'answerdate_month': 'April',
            'answerdate_day': '04',
            'answerdate_hour': '09',
            'answerdate_minute': '00'}).submit()

        self.assertEqual(
            ['No response text added and no issue changes made.'],
            statusmessages.error_messages())

    @browsing
    def test_do_not_update_anserdate_if_no_month_is_given(self, browser):
        ticketbox = create(Builder('ticket box'))
        ticket = create(Builder('ticket')
                        .having(answerDate=DateTime(2015, 12, 31))
                        .within(ticketbox))

        browser.login().visit(ticket)

        browser.fill({
            'answerdate_year': '2021',
            'answerdate_month': '--',
            'answerdate_day': '04',
            'answerdate_hour': '09',
            'answerdate_minute': '00'}).submit()

        self.assertEqual(
            ['No response text added and no issue changes made.'],
            statusmessages.error_messages())

    @browsing
    def test_do_not_update_anserdate_if_no_day_is_given(self, browser):
        ticketbox = create(Builder('ticket box'))
        ticket = create(Builder('ticket')
                        .having(answerDate=DateTime(2015, 12, 31))
                        .within(ticketbox))

        browser.login().visit(ticket)

        browser.fill({
            'answerdate_year': '2021',
            'answerdate_month': 'April',
            'answerdate_day': '--',
            'answerdate_hour': '09',
            'answerdate_minute': '00'}).submit()

        self.assertEqual(
            ['No response text added and no issue changes made.'],
            statusmessages.error_messages())

    @browsing
    def test_update_anserdate_if_no_hour_or_minute_is_given(self, browser):
        ticketbox = create(Builder('ticket box'))
        ticket = create(Builder('ticket')
                        .having(answerDate=DateTime(2015, 12, 31))
                        .within(ticketbox))

        browser.login().visit(ticket)

        browser.fill({
            'answerdate_year': '2021',
            'answerdate_month': 'April',
            'answerdate_day': '12',
            'answerdate_hour': '--',
            'answerdate_minute': '--'}).submit()

        self.assertEqual(
            1, len(browser.css('.response-reply')),
            "There should be one response visible")



from setuptools import setup, find_packages
import os

version = '4.9.0'
maintainer = 'Mathias Leimgruber'

tests_require = [
    'unittest2',
    'ftw.testing',
    'plone.app.testing',
    'ftw.builder',
    'ftw.testbrowser',
    'ftw.zipexport',
    'collective.MockMailHost',
    'xlrd',
]

setup(
    name='izug.ticketbox',
    version=version,
    description="A tracker-like task management system for plone.",
    long_description=(open('README.rst').read() + '\n' + \
                      open(os.path.join('docs', 'HISTORY.txt')).read()),

    # Get more strings from
    # http://www.python.org/pypi?%3Aaction=list_classifiers

    classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.2',
        'Framework :: Plone :: 4.3',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],

    keywords='ticketbox tracker ftw ',
    author='4teamwork AG',
    author_email='mailto:info@4teamwork.ch',
    maintainer=maintainer,
    url='https://github.com/4teamwork/izug.ticketbox',
    license='GPL2',

    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['izug', ],
    include_package_data=True,
    zip_safe=False,

    install_requires=[
        'setuptools',
        'AccessControl',
        'Acquisition',
        'Plone',
        'Products.ATContentTypes',
        'Products.ATReferenceBrowserWidget',
        'Products.Archetypes',
        'Products.CMFCore',
        'Products.CMFPlone',
        'Products.DataGridField',
        'Products.statusmessages',
        'ZODB3',
        'Zope2',
        'ftw.calendarwidget',
        'ftw.notification.base',
        'ftw.notification.email',
        'ftw.tabbedview',
        'ftw.table',
        'ftw.upgrade >= 1.14.0',
        'plone.principalsource',
        'zope.annotation',
        'zope.app.component',
        'zope.app.container',
        'zope.app.pagetemplate',
        'zope.cachedescriptors',
        'zope.component',
        'zope.contentprovider',
        'zope.event',
        'zope.i18n',
        'zope.i18nmessageid',
        'zope.interface',
        'zope.lifecycleevent',
        'zope.schema',
        'zope.viewlet',
        'XlsxWriter',
    ],

    tests_require=tests_require,
    extras_require=dict(tests=tests_require,
                        workspace=['ftw.workspace'],
                        zip_export=['ftw.zipexport']),

    test_suite='izug.ticketbox.tests.test_docs.test_suite',
    entry_points="""
    # -*- entry_points -*-
    [z3c.autoinclude.plugin]
    target = plone
    """,
)



from ftw.upgrade import UpgradeStep


class InstallFtwCalendarWidget(UpgradeStep):
    """Install ftw calendar widget.
    """

    def __call__(self):
        self.setup_install_profile('profile-ftw.calendarwidget:default')
